"""VRE Representation module"""
from __future__ import annotations
from abc import ABC, abstractmethod
import numpy as np
from vre_video import VREVideo

from ..utils import parsed_str_type, ReprOut
from ..logger import vre_logger as logger

class Representation(ABC):
    """Generic Representation class for VRE"""
    def __init__(self, name: str, dependencies: list[Representation] | None = None):
        deps = [] if dependencies is None else dependencies
        assert isinstance(deps, list), type(deps)
        self.name = name
        self.dependencies = deps
        self._batch_size: int | None = None

    @abstractmethod
    def make_images(self, data: ReprOut) -> np.ndarray:
        """Given the output of self.compute(video, ixs) of type ReprOut, return a [0:255] image for each frame"""

    @property
    @abstractmethod
    def n_channels(self) -> int:
        """The number of channels (dimensions, not resolution) of this representation"""

    ## Public properties & methods

    @property
    def batch_size(self) -> int:
        """the batch size that is used during the VRE run for computation"""
        if self._batch_size is None:
            logger.debug(f"[{self}] No batch_size set, returning 1. Call set_compute_params")
            return 1
        return self._batch_size

    @batch_size.setter
    def batch_size(self, bs: int):
        assert isinstance(bs, int) and bs >= 1, type(bs)
        self._batch_size = bs

    @property
    def dep_names(self) -> list[str]:
        """return the list of dependencies names"""
        return [r.name for r in self.dependencies]

    @property
    def is_classification(self) -> bool:
        """if we have self.classes. Used in MultiTaskReader."""
        return hasattr(self, "classes") and self.classes is not None # pylint: disable=no-member

    def size(self, repr_out: ReprOut) -> tuple[int, ...]:
        """Returns the (b, h, w, c) tuple of the size of the current representation"""
        return tuple(repr_out.output.shape)

    def compute(self, video: VREVideo, ixs: list[int], dep_data: list[ReprOut] | None = None) -> ReprOut:
        """
        Given input data (batch of images/frewes), compute the output data of this representation.
        Not abstract so we can instantiate testing representations and enable syntax highligting w/o CompputeReprMixin.
        """
        raise NotImplementedError(f"{self} must implement .compute()")

    def set_compute_params(self, **kwargs):
        """set the compute parameters for the representation"""
        attributes = ["batch_size"]
        res = ""
        assert set(kwargs).issubset(attributes), (list(kwargs), attributes)
        for attr in attributes:
            if attr in kwargs:
                setattr(self, attr, kwargs[attr])
                res += f"\n-{attr}: {kwargs[attr]}"
        if len(res) > 0:
            logger.debug(f"[{self}] Set node specific 'Compute' params:{res}")

    ## Magic methods ##
    def __repr__(self):
        return f"{parsed_str_type(self)}({self.name}{f' {self.dep_names}' if len(self.dep_names) > 0 else ''})"



"""Metadata module -- The metadata of a particular VRE run"""
from __future__ import annotations
import json
from typing import Any
from pathlib import Path

from .representation_metadata import RepresentationMetadata
from .utils import random_chars, mean, now_fmt
from .vre_runtime_args import VRERuntimeArgs

class RunMetadata:
    """Metadata of a run for multiple representations. Backed on the disk by a JSON file"""
    def __init__(self, repr_names: list[str], runtime_args: VRERuntimeArgs, logs_dir: Path,
                 now_str: str | None = None, run_id: str | None = None):
        assert len(repr_names) > 0 and all(isinstance(x, str) for x in repr_names), repr_names
        assert (A := set(r.name for r in runtime_args.representations)) == (B := set(repr_names)), (A, B)
        self.repr_names = repr_names
        self.logs_dir = logs_dir
        self.runtime_args = runtime_args.to_dict()
        self.id = run_id or random_chars(n=10)
        self.data_writers = {}
        self.run_stats = {}
        self.total_stats = {
            "n_computed": 0,
            "n_failed": 0,
            "duration": 0,
        }
        self.now_str = now_str or now_fmt()
        self.store_on_disk()

    @property
    def disk_location(self) -> Path:
        """The full path disk location of this metadata"""
        return self.logs_dir / f"run_metadata-{self.id}-{self.now_str}.json"

    @property
    def metadata(self) -> dict[str, Any]:
        """the run metadata as a json"""
        return {
            "id": self.id,
            "runtime_args": self.runtime_args,
            "data_writers": self.data_writers,
            "run_stats": self.run_stats,
            "total": self.total_stats,
        }

    def add_run_stats(self, repr_metadata: RepresentationMetadata):
        """adds statistics of a single representation after it finished running"""
        assert (name := repr_metadata.repr_name) not in self.run_stats, f"{name} in {self.run_stats.keys()}"
        frames_computed = repr_metadata.frames_computed(run_id=self.id)
        frames_failed = repr_metadata.frames_failed(run_id=self.id)

        avg_duration = round(mean([repr_metadata.run_stats[ix].duration for ix in frames_computed]), 3)
        self.run_stats[name] = {"n_computed": len(frames_computed), "n_failed": len(frames_failed),
                                "average_duration": avg_duration}
        self.total_stats = {
            "n_computed": self.total_stats["n_computed"] + len(frames_computed),
            "n_failed": self.total_stats["n_failed"] + len(frames_failed),
            "duration": round(self.total_stats["duration"] + avg_duration * len(frames_computed), 3)
        }
        self.store_on_disk()

    def store_on_disk(self):
        """stores (overwrites if needed) the metadata on disk"""
        with open(self.disk_location, "w") as fp:
            json.dump(self.metadata, fp, indent=4)

    def __repr__(self):
        return (f"[Metadata] Id: {self.id}, Num representations: {len(self.repr_names)}. "
                f"Disk location: '{self.disk_location}'")



"""fetch.py -- The VRE files fetcher"""
from pathlib import Path
from urllib.request import urlretrieve
from tqdm import tqdm

from ..logger import vre_logger as logger

class DownloadProgressBar(tqdm):
    """requests + tqdm"""
    def update_to(self, b=1, bsize=1, tsize=None):
        """Callback from tqdm"""
        if tsize is not None:
            self.total = tsize
        self.update(b * bsize - self.n)

def is_git_lfs(path: Path) -> bool:
    """Returns true if a path is a git lfs link"""
    if path.is_dir():
        return False
    with open(path, "rb") as fp:
        try:
            return fp.read(7).decode("utf-8") == "version"
        except UnicodeDecodeError:
            return False

def fetch(src: str, dst: Path) -> Path:
    """fetches the source url to a destionat path. Returns that path. Support for git lfs and recursive mkdir()"""
    if not dst.exists() or is_git_lfs(dst):
        dst.parent.mkdir(exist_ok=True, parents=True)
        with DownloadProgressBar(unit="B", unit_scale=True, miniters=1, desc=dst.name) as t:
            try:
                urlretrieve(src, filename=dst, reporthook=t.update_to)
            except Exception as e:
                logger.info(f"Failed to download '{src}' to '{dst}'")
                raise e
    return dst



"""Init file"""
# pylint: disable=reimported, wrong-import-position
import warnings
try:
    from diffusers.utils.logging import disable_progress_bar
    disable_progress_bar() # pylint: disable=all
except ImportError:
    pass


from .video_representations_extractor import VideoRepresentationsExtractor, VideoRepresentationsExtractor as VRE
from .representations import Representation, ReprOut
from .utils.lovely import monkey_patch
from .utils.repr_memory_layout import MemoryData, DiskData

try:
    import pdbp
except ImportError:
    pass

monkey_patch()

try:
    import torch # noqa: F401
    warnings.filterwarnings("ignore", "You are using `torch.load` with `weights_only=False`*.")
    warnings.filterwarnings("ignore", "`clean_up_tokenization_spaces` was not set*.")
except ImportError:
    pass



"""Video Repentations Extractor module"""
from __future__ import annotations
from pathlib import Path
from datetime import datetime
import os
import traceback
from tqdm import tqdm
import numpy as np
from vre_video import VREVideo

from .representations import Representation, LearnedRepresentationMixin, RepresentationsList, IORepresentationMixin
from .vre_runtime_args import VRERuntimeArgs
from .data_writer import DataWriter
from .data_storer import DataStorer
from .run_metadata import RunMetadata
from .representation_metadata import RepresentationMetadata
from .utils import now_fmt, make_batches, ReprOut, DiskData, SummaryPrinter, random_chars, MemoryData
from .logger import vre_logger as logger

# TODO: split in 2 classes ?
# vre_batched = VREBatched(VREVideo, RepresentationsList); vre_batch.run(...)
# vre_streaming = VREStreaming(VREVIdeo(??) RepresentationsList); vre_streaming[0:5]

class VideoRepresentationsExtractor:
    """Video Representations Extractor class"""

    def __init__(self, video: VREVideo, representations: RepresentationsList | list[Representation]):
        """
        Parameters:
        - video The video we are performing VRE one
        - representations The list instantiated representations. Must be topo-sortable based on name and deps.
        """
        assert isinstance(video, VREVideo), (type(video), video)
        assert isinstance(representations, (list, tuple, RepresentationsList)), type(representations)
        self.video = video
        self.representations = RepresentationsList(representations)

    def set_compute_params(self, **kwargs) -> VideoRepresentationsExtractor:
        """Set the required params for all representations"""
        for r in self.representations:
            r.set_compute_params(**kwargs)
        return self

    def set_io_parameters(self, **kwargs) -> VideoRepresentationsExtractor:
        """Set the required params for all representations of IORepresentationMixin type"""
        for r in [_r for _r in self.representations if isinstance(_r, IORepresentationMixin)]:
            r.set_io_params(**kwargs)
        return self

    def run(self, output_dir: Path, frames: list[int] | None = None, output_dir_exists_mode: str = "raise",
            exception_mode: str = "stop_execution", n_threads_data_storer: int = 0,
            subset_exported_representations: list[str] | None = None) -> RunMetadata:
        """
        The main loop of the VRE. This will run all the representations on the video and store results in the output_dir
        Parameters:
        - output_dir The directory where VRE will store the representations
        - frames The list of frames to run VRE for. If None, will export all the frames of the video
        - output_dir_exists_mode What to do if the output dir already exists. See DataWriter docstring. Default 'raise'.
        - exception_mode What to do when encountering an exception. It always writes the exception to 'exception.txt'
          - 'skip_representation' Will stop the run of the current representation and start the next one
          - 'stop_execution' (default) Will stop the execution of VRE
        - n_threads_data_storer The number of threads used by the DataStorer
        - subset_exported_representations If set, only this subset of representations are exported, otherwise all of
            them based on these provided to the VRE constructor.
        Returns:
        - A RunMetadata object representing the run statistics for all representations of this run.
        """
        self._setup_logger(logs_dir := output_dir / ".logs", run_id := random_chars(n=10), now := now_fmt())
        self._setup_graphviz(logs_dir, run_id, now)
        if (subset := subset_exported_representations) is not None:
            logger.info(f"Explicit subset provided: {subset}. Exporting only these.")

        exported_reprs = self.representations.get_output_representations(subset=subset)
        assert len(exported_reprs) > 0, f"No output reprs returned, set I/O! {self.representations=}, {subset=}"
        runtime_args = VRERuntimeArgs(video=self.video, representations=exported_reprs, frames=frames,
                                      exception_mode=exception_mode, n_threads_data_storer=n_threads_data_storer)
        run_metadata = RunMetadata(repr_names=exported_reprs.names, runtime_args=runtime_args,
                                   logs_dir=logs_dir, now_str=now, run_id=run_id)
        logger.info(runtime_args)
        summary_printer = SummaryPrinter(exported_reprs.names, runtime_args)

        for vrepr in exported_reprs:
            dw = DataWriter(output_dir=output_dir, representation=vrepr, output_dir_exists_mode=output_dir_exists_mode)
            run_metadata.data_writers[vrepr.name] = dw.to_dict()
            repr_metadata = self.do_one_representation(run_id=run_metadata.id, representation=vrepr,
                                                       output_dir=output_dir,
                                                       output_dir_exists_mode=output_dir_exists_mode,
                                                       runtime_args=runtime_args)
            if repr_metadata.run_had_exceptions and runtime_args.exception_mode == "stop_execution":
                raise RuntimeError(f"Representation '{vrepr.name}' threw. "
                                   f"Check '{logger.get_file_handler().file_path}' for information")
            run_metadata.add_run_stats(repr_metadata)
            summary_printer.repr_metadatas[vrepr.name] = repr_metadata
        print(summary_printer())
        return run_metadata

    def do_one_representation(self, run_id: str, representation: Representation, output_dir: Path,
                              output_dir_exists_mode: str, runtime_args: VRERuntimeArgs) -> RepresentationMetadata:
        """The loop of each representation. Returns a representation metadata with information about this repr's run"""
        data_writer = DataWriter(output_dir=output_dir, representation=representation,
                                 output_dir_exists_mode=output_dir_exists_mode)
        data_storer = DataStorer(data_writer=data_writer, n_threads=runtime_args.n_threads_data_storer)
        logger.debug(f"Running {run_id=}:\n{representation}\n{data_storer}")
        rep: Representation | IORepresentationMixin = representation
        formats = sorted([f for f in [rep.image_format.value, rep.binary_format.value] if f != "not-set"])
        repr_metadata = RepresentationMetadata(repr_name=representation.name, formats=formats,
                                               disk_location=data_writer.rep_out_dir / ".repr_metadata.json",
                                               frames=list(range(len(self.video))))

        relevant_frames = [f for f in runtime_args.frames if f not in repr_metadata.frames_computed()]
        logger.debug(f"Out of {len(runtime_args.frames)} total frames, "
                     f"{len(runtime_args.frames) - len(relevant_frames)} are precomputed and will be skipped.")
        batches = make_batches(relevant_frames, rep.batch_size)

        pbar = tqdm(total=len(relevant_frames), desc=f"[VRE] {rep.name} bs={rep.batch_size}",
                    disable=os.getenv("VRE_PBAR", "1") == "0")
        for batch in batches:
            now = datetime.now()
            if data_writer.all_batch_exists(batch):
                pbar.update(len(batch))
                continue

            try:
                # tr.cuda.empty_cache()?
                rep_data = self._compute_one_representation_batch(rep, batch=batch, output_dir=data_writer.output_dir)
                if not rep.is_classification and rep.name.find("fastsam") == -1: # TODO: MemoryData of FSAM is binary...
                    assert rep_data.output.shape[-1] == rep.n_channels, (rep, rep_data.output, rep.n_channels)
                rep_data.output_images = rep.make_images(rep_data) if rep.export_image else None
                data_storer(rep_data)
            except Exception:
                self._log_error(f"\n[{rep.name} {rep.batch_size=} {batch=}] {traceback.format_exc()}\n")
                repr_metadata.add_time(None, batch, run_id=run_id, sync=True)
                repr_metadata.run_had_exceptions = True
                break
            repr_metadata.add_time((datetime.now() - now).total_seconds(), batch, run_id=run_id, sync=True)
            pbar.update(len(batch))

        # various cleanup stuff before ending with this representation
        data_storer.join_with_timeout(timeout=30)
        if isinstance(representation, LearnedRepresentationMixin) and representation.setup_called:
            representation.vre_free()
        for dep in representation.dependencies:
            if isinstance(dep, LearnedRepresentationMixin) and dep.setup_called:
                dep.vre_free()
        repr_metadata.store_on_disk()
        return repr_metadata

    # Helper private methods

    def _load_from_disk_if_possible(self, rep: Representation, video: VREVideo, ixs: list[int],
                                    output_dir: Path) -> ReprOut | None:
        """loads (batched) data from disk if possible."""
        assert isinstance(rep, IORepresentationMixin), rep
        assert isinstance(ixs, list) and all(isinstance(ix, int) for ix in ixs), (type(ixs), [type(ix) for ix in ixs])
        assert output_dir is not None and output_dir.exists(), output_dir
        # TODO: use data writer to create the paths as they can be non-npz for example
        npz_paths: list[Path] = [output_dir / rep.name / f"npz/{ix}.npz" for ix in ixs]
        extra_paths: list[Path] = [output_dir / rep.name / f"npz/{ix}_extra.npz" for ix in ixs]
        if any(not x.exists() for x in npz_paths): # partial batches are considered 'not existing' and overwritten
            return None
        extras_exist = [x.exists() for x in extra_paths]
        assert (ee := sum(extras_exist)) in (0, (ep := len(extra_paths))), f"Found {ee}. Expected either 0 or {ep}"
        disk_data: DiskData = np.array([rep.load_from_disk(x) for x in npz_paths])
        extra = [np.load(x, allow_pickle=True)["arr_0"].item() for x in extra_paths] if ee == ep else None
        logger.debug2(f"[{rep}] Slice: [{ixs[0]}:{ixs[-1]}]. All data found on disk and loaded")
        return ReprOut(frames=video[ixs], output=rep.disk_to_memory_fmt(disk_data), extra=extra, key=ixs)

    def _compute_one_representation_batch(self, rep: Representation, batch: list[int],
                                          output_dir: Path, depth: int = 0) -> ReprOut:
        # setting depth to higher values allows to compute deps of deps in memory. We throw to not hide bugs.
        # useful for VRE hacking (i.e. see semantic_mapper.py)
        assert depth <= 1, f"{rep=} {depth=} {batch=} {output_dir=}"
        loaded_data = self._load_from_disk_if_possible(rep, self.video, batch, output_dir)
        if loaded_data is not None:
            return loaded_data
        dep_data = []
        for dep in rep.dependencies:
            dep_data.append(self._compute_one_representation_batch(dep, batch, output_dir, depth + 1))
        if isinstance(rep, LearnedRepresentationMixin) and rep.setup_called is False:
            rep.vre_setup() # instantiates the model, loads to cuda device etc.
        return rep.compute(self.video, ixs=batch, dep_data=dep_data)

    def _setup_graphviz(self, logs_dir: Path, run_id: str, now_str: str):
        try:
            self.representations.to_graphviz().render(pth := f"{logs_dir}/graph-{run_id}-{now_str}",
                                                      format="png", cleanup=True)
            logger.info(f"Stored graphviz representation at: '{pth}.png'")
        except Exception as e:
            logger.error(e)

    def _setup_logger(self, logs_dir: Path, run_id: str, now_str: str):
        logs_dir.mkdir(exist_ok=True, parents=True)
        logs_file = logs_dir / f"logs-{run_id}-{now_str}.txt"
        logger.add_file_handler(logs_file)
        logger.info(f"Logging run at: '{logs_file}")

    def _log_error(self, msg: str):
        logs_file = logger.get_file_handler().file_path
        open(logs_file, "a").write(f"{'=' * 80}\n{now_fmt()}\n{msg}\n{'=' * 80}")
        logger.debug(f"Error: {msg}")

    def __call__(self, *args, **kwargs) -> RunMetadata:
        return self.run(*args, **kwargs)

    def __str__(self) -> str:
        return f"""\n[VRE]
- Video: {self.video}
- Representations: {self.representations}"""

    def __repr__(self) -> str:
        return str(self)

    def __getitem__(self, ix: int | slice | range | list[int]) -> dict[str, ReprOut]:
        # Note: setup and free should be called outside of this function!
        if isinstance(ix, int):
            return self[[ix]]
        if isinstance(ix, slice):
            return self[list(range(ix.start, ix.stop))]
        if isinstance(ix, range):
            return self[list(ix)]
        assert isinstance(ix, list), type(ix)
        assert all(isinstance(r, IORepresentationMixin) for r in self.representations), self.representations

        res: dict[str, ReprOut] = {}
        for vre_repr in (pbar := tqdm(self.representations, disable=os.getenv("VRE_PBAR", "1") == "0")):
            pbar.set_description(f"[VRE Streaming] {vre_repr.name}")
            dep_names = [r.name for r in vre_repr.dependencies]
            batches = make_batches(ix, vre_repr.batch_size)
            batch_res: list[ReprOut] = []
            for b_ixs in batches:
                repr_out = vre_repr.compute(video=self.video, ixs=b_ixs,
                                            dep_data=[res[dep_name] for dep_name in dep_names])
                if vre_repr.output_size is not None:
                    repr_out = vre_repr.resize(repr_out, vre_repr.output_size)
                batch_res.append(repr_out)
            combined = ReprOut(frames=np.concatenate([br.frames for br in batch_res]),
                               key=sum([br.key for br in batch_res], []),
                               output=MemoryData(np.concatenate([br.output for br in batch_res])),
                               extra=sum([br.extra if br.extra is not None else [] for br in batch_res], []))
            if vre_repr.image_format.value != "not-set":
                combined.output_images = vre_repr.make_images(combined)
            res[vre_repr.name] = combined
        return res



"""Resources utils. Stuff related to the resources stored in the repo via git-lfs. Weights are not in resources!"""
from pathlib import Path
import os

from .utils import get_project_root
from .fetch import fetch

REPO_URL = "https://gitlab.com/video-representations-extractor/video-representations-extractor"
RESOURCES_URL = f"{REPO_URL}/-/raw/master/resources"
RESOURCES_DIR = Path(os.getenv("VRE_RESOURCES_DIR", get_project_root() / "resources"))

def fetch_resource(resource_name: str) -> Path:
    """fetches a resources from gitlab LFS if needed"""
    return fetch(src=f"{RESOURCES_URL}/{resource_name}", dst=RESOURCES_DIR / resource_name)



"""RepresentationMetadata -- the metadata of a representation stored on disk"""
from pathlib import Path
from typing import Any, NamedTuple
from io import FileIO
import json

from .utils import AtomicOpen

FrameStats = NamedTuple("frame_stats", duration=float, run_id=str, formats=list[str])

class RepresentationMetadata:
    """
    A class that defines the metadata and events that happened during a single representation's run.
    It is backed by a single sqlite file living in the representation's directory under .run_metadata.json.
    Note: that this file may be updated from multiple processes running at the same time. For this reason, we have an
    extra layer protection based on json modification time via `os.path.getmtime` and we merge before storing to disk.
    """
    def __init__(self, repr_name: str, disk_location: Path, frames: list[int], formats: list[str]):
        assert all(isinstance(x, int) for x in frames), frames
        assert len(formats) > 0, "formats cannot be empty, is your representation an IORepresentation?"
        self.repr_name = repr_name
        self.disk_location = disk_location
        self.run_had_exceptions = False
        self.formats = formats
        self.run_stats: dict[int, FrameStats | None] = {f: None for f in frames}
        self.store_on_disk()

    def frames_computed(self, run_id: str | None = None) -> list[int]:
        """returns the list of comptued frames so far for this representation"""
        def condition(frame_stats: FrameStats | None) -> bool:
            if frame_stats is None or frame_stats.duration is None:
                return False
            return set(self.formats).issubset(frame_stats.formats)
        return [ix for ix, v in self.run_stats.items() if condition(v)
                and (v.run_id == run_id if run_id is not None else True)]

    def frames_failed(self, run_id: str | None = None) -> list[int]:
        """returns the list of failed frames so far for this representation"""
        return [ix for ix, v in self.run_stats.items() if v is not None and v.duration is None
                and (v.run_id == run_id if run_id is not None else True)]

    def add_time(self, duration: float | None, frames: list[int], run_id: str, sync: bool=False):
        """adds a (batched) time to the representation's run_stats. If sync is true, it also calls store_on_disk."""
        assert (batch_size := len(frames)) > 0, batch_size
        data = [duration / batch_size] * batch_size if duration is not None else [None] * batch_size
        for ix, frame_duration in zip(frames, data):
            if self.run_stats[ix] is not None and self.run_stats[ix].duration is not None:
                if set(self.formats).issubset(self.run_stats[ix].formats):
                    raise ValueError(f"Adding time to existing metadata {self}. Frame={ix}. "
                                     f"Previous: {self.run_stats[ix]}")
                # overwrite the run id but add the previous time of the old representations to current frame duration
                frame_duration += self.run_stats[ix].duration
            self.run_stats[ix] = FrameStats(run_id=run_id, duration=frame_duration, formats=self.formats)
        if sync:
            self.store_on_disk()

    def to_dict(self, sync: bool=False) -> dict:
        """returns the metadata as a dict. Used mostly for tests. If atomic is set to true, uses AtomicOpen"""
        if sync and self.disk_location.exists():
            with AtomicOpen(self.disk_location, "a+") as fp:
                self.run_stats = self._load_and_merge_with_metadata_from_disk(fp)
                fp.seek(0)
                fp.truncate()

        return {
            "name": self.repr_name,
            "run_stats": {k: None if v is None else v._asdict() for k, v in self.run_stats.items()},
        }

    def store_on_disk(self):
        """
        Stores (overwrites) the metadata on disk. Note: this does allows for multi processing on the same repr!
        Note: that disk_writer_meta will be overwritten by the last writer.
        """
        file_exists = self.disk_location.exists()
        with AtomicOpen(self.disk_location, "a+") as fp:
            if file_exists: # if it exists, we merge the disk data with existing data before overwriting
                self.run_stats = self._load_and_merge_with_metadata_from_disk(fp)

            fp.seek(0)
            fp.truncate()
            json_data = {
                "name": self.repr_name,
                "run_stats": {k: None if v is None else v._asdict() for k, v in self.run_stats.items()},
            }
            json.dump(json_data, fp, indent=4)

    def _load_and_merge_with_metadata_from_disk(self, fp: FileIO) -> dict[str, Any]:
        """given an (mutexed) fp, read the metadata from the disk and return the run stats. Fail if any errors"""
        fp.seek(0)
        raw_data = fp.read()
        loaded_json_data = json.loads(raw_data)
        loaded_run_stats = {int(k): None if v is None else FrameStats(**v)
                            for k, v in loaded_json_data["run_stats"].items()}
        assert (a := loaded_run_stats.keys()) == (b := self.run_stats.keys()), f"\n- {a}\n- {b}"
        assert (a := loaded_json_data["name"]) == (b := self.repr_name), f"\n- {a}\n- {b}"
        merged_run_stats = {}
        # TODO... test this stuff below
        for k, v in self.run_stats.items():
            if (loaded_v := loaded_run_stats[k]) is not None and loaded_v.duration is not None:
                if v is None:
                    merged_run_stats[k] = loaded_v
                    continue
                # v exists too below
                if set(self.formats) == set(loaded_v.formats) and v != loaded_v:
                    raise ValueError(k, v, loaded_v)
                # v exists and is different from loaded_v (i.e. jpg in new run and npz in prev) -> merge formats
                merged_formats = sorted(set(v.formats).union(loaded_v.formats))
                merged_run_stats[k] = FrameStats(duration=v.duration, formats=merged_formats, run_id=v.run_id)
            else:
                merged_run_stats[k] = v
        return merged_run_stats

    def __repr__(self):
        return (f"[ReprMetadata] Representation: {self.repr_name}. Frames: {len(self.run_stats)} "
                f"(computed: {len(self.frames_computed())}, failed: {len(self.frames_failed())}) "
                f"Disk location: '{self.disk_location}'")



"""Weights Repository module for Learnable Representations in the default repository"""
from pathlib import Path

import torch as tr
from vre.utils import fetch
from vre.logger import vre_logger as logger

REPO_URL = "https://gitlab.com/video-representations-extractor/video-representations-extractor"
VRE_REPO_URL = f"{REPO_URL}/-/raw/master/vre_repository"

def fetch_weights(paths: list[str | Path]) -> list[Path]:
    """
    Fetches weights for a representation from the repository (if needed) and returns the local path.
    The paths must be of the format: "/path/to/vre_repository/..representation../weights/ckpt(s)".
    """
    assert isinstance(paths, list), (paths, type(paths))
    assert all(isinstance(p, (str, Path, list)) for p in paths), paths
    assert all(str(x).find("/vre_repository/") for x in paths), paths
    res = []
    for path in paths:
        src = VRE_REPO_URL + str(path)[str(path).index("vre_repository")+14:]
        if isinstance(path, list):
            res.append(fetch_weights(path))
        else:
            res.append(fetch(src, path))
    return res

def vre_load_weights(path: Path | list[Path | str]) -> dict[str, tr.Tensor]:
    """load weights from disk. weights can be sharded as well. Sometimes we store this due to git-lfs big files bug."""
    logger.debug(f"Loading weights from '{path}'")

    if isinstance(path, (str, Path)) and path.is_dir():
        path = list(path.iterdir())

    if isinstance(path, list):
        res = {}
        for item in path:
            res = {**res, **tr.load(item, map_location="cpu")}
    else:
        res = tr.load(path, map_location="cpu")
    return res



"""representations_list.py: module that implements a wrapper on top of a list of representations (auto sorting etc.)"""
from __future__ import annotations
from typing import Any, Iterator
from overrides import overrides
from vre.utils import vre_topo_sort
from .representation import Representation
from .io_representation_mixin import IORepresentationMixin

class RepresentationsList(list):
    """module that topo-sorts and implements helper functions on top of a representation list"""
    def __init__(self, representations: list[Representation] | RepresentationsList, topo_sort: bool = True):
        representations = vre_topo_sort(representations) if topo_sort else representations
        super().__init__(representations)
        assert len(self) > 0, "At least one representation must be provided"
        assert all(isinstance(x, Representation) for x in self), [(x.name, type(x)) for x in self]
        self.names = [r.name for r in self]

    @overrides
    def append(self, object: Any): # pylint: disable=redefined-builtin
        raise ValueError("Cannot append, use the constructor")

    def get_output_representations(self, subset: list[str] | None = None) -> RepresentationsList:
        """
        Given all the representations, keep those that actually export something. RunMetadata uses only these.
        We do a bunch of checks:
        - at least one representation is provided
        - they are exportable: IORpresentations are set, like npz, png etc.
        - no duplicates provided
        Returns a subset of representations from the ones given at constructor that will be exported.
        """
        subset = subset or self.names
        assert all(name in self.names for name in subset), (
            f"{subset=}\n{self.names=}\nmissing={[n for n in subset if n not in self.names]}")

        io_reprs: list[IORepresentationMixin] = [_r for _r in self if isinstance(_r, IORepresentationMixin)]
        res = []
        for r in io_reprs:
            if (r.export_binary or r.export_image) and r.name in subset:
                res.append(r)
        return RepresentationsList(res, topo_sort=False)

    def to_graphviz(self, **kwargs) -> "Digraph":
        """Returns a graphviz object from this graph. Used for plotting the graph. Best for smaller graphs."""
        from graphviz import Digraph # pylint: disable=import-outside-toplevel
        g = Digraph()
        for k, v in kwargs.items():
            g.attr(**{k: v})
        g.attr(rankdir="LR")
        for node in self:
            g.node(name=f"{node.name}", shape="oval")
        edges: list[tuple[str, str]] = [(r.name, dep.name) for r in self for dep in r.dependencies]
        for l, r in edges:
            g.edge(r, l) # reverse?
        return g

    @overrides
    def __iter__(self) -> Iterator[Representation]: # for type-hinting
        return super().__iter__()

    def __str__(self) -> str:
        return f"Representations ({len(self)}): [{', '.join(self.names)}]"

    def __repr__(self):
        return str(self)



"""RGB representation."""
from __future__ import annotations
from overrides import overrides

from vre_video import VREVideo
from vre.utils import MemoryData, ReprOut

from .color_representation import ColorRepresentation

class RGB(ColorRepresentation):
    """Basic RGB representation."""
    def __init__(self, *args, **kwargs):
        ColorRepresentation.__init__(self, *args, **kwargs)
        assert len(self.dependencies) == 0, self.dependencies
        self.output_dtype = "uint8"

    @overrides
    def compute(self, video: VREVideo, ixs: list[int], dep_data: list[ReprOut] | None = None) -> ReprOut:
        return ReprOut(frames=video[ixs], output=MemoryData(video[ixs]), key=ixs) # video[ixs] is cached



"""
yaml utils -- a minimum subset of stuff from omegaconf
Warning! This yaml syntax is not supported:
key: {other_key: ${oc.env}}
Use this instead:
key:
  other_key: ${oc.env}
"""
from io import IOBase, StringIO
from pathlib import Path
import ast
import os
from typing import Any
import yaml

def _parse(data: str) -> str:
    if (ix := data.find("${")) == -1:
        return data
    left, current = data[0:ix], data[ix:]
    if current.startswith("${oc.env:"):
        current = _parse(current[len("${oc.env:"):]) # recursively resolve inner oc.env:{} stuff
        assert (ix_r := current.find("}")) != -1, current
        key_and_default = current[0:ix_r].split(",")
        assert len(key_and_default) >= 1, current
        key, default = key_and_default[0], None
        if len(key_and_default):
            default = ",".join(key_and_default[1:]) # in case we have list defaults
        if len(key_and_default) == 1 and key not in os.environ:
            raise KeyError(f"Environment variable: '{key}' not set and no default provided.")
        value = os.getenv(key, default)
        current = f"{value}{current[ix_r+1:]}" # get rid of the '}' of this oc.env and replace it with the env value
    if current.startswith("${oc.decode:"):
        current = _parse(current[len("${oc.decode:"):]) # recursively resolve inner oc.env:{} stuff
        assert (ix_r := current.find("}")) != -1, current
        value = ast.literal_eval(current[0:ix_r])
        current = f"{value}{current[ix_r+1:]}" # get rid of the '}' of this oc.decode and replace it with the decoded
    return f"{left}{current}"

def vre_yaml_load(path: Path | str | IOBase) -> dict[str, Any]:
    """reads a yaml file and resolves the env varaibles if needed"""
    fp = open(path, "r") if isinstance(path, (Path, str)) else path
    fp.seek(0)
    data: str = fp.read()
    assert len(data) > 0, f"Nothing read from '{path}'"

    cfg_str = _parse(data)
    try:
        cfg = yaml.safe_load(StringIO(cfg_str))
    except Exception as e:
        print(f"-Could not parse:\n{cfg_str}\n-Exception: {e}")
        raise e
    fp.close() if isinstance(path, (Path, str)) else None
    return cfg



"""cv2 utils. All the calls to opencv must be condensed here"""
import numpy as np
from PIL import Image, ImageDraw, ImageOps, ImageFont

from .utils import get_project_root
from ..logger import vre_logger as logger

def pil_image_resize(data: np.ndarray, height: int, width: int, interpolation: str, **kwargs) -> np.ndarray:
    """Wrapper on top of Image(arr).resize((w, h), args)"""
    interpolation = {"nearest": Image.Resampling.NEAREST, "bilinear": Image.Resampling.BILINEAR}[interpolation]
    if data.dtype == np.uint8:
        return np.asarray(Image.fromarray(data).resize((width, height), interpolation, **kwargs))
    # For non-uint8 (e.g. float32), resize each channel via PIL mode "F" and reassemble
    assert np.issubdtype(data.dtype, np.floating), f"Only uint8 or float allowed, got {data.dtype}"
    channels = [data[..., c] for c in range(data.shape[-1])] if data.ndim == 3 else [data]
    resized = [np.asarray(Image.fromarray(ch.astype(np.float32), mode="F").resize((width, height), interpolation, **kwargs))
               for ch in channels]
    return np.stack(resized, axis=-1).astype(data.dtype) if data.ndim == 3 else resized[0].astype(data.dtype)

def pil_image_add_title(image: np.ndarray, text: str, font: str = None, font_color: str = "white",
                        background_color: str = "black", top_padding: int = None, size_px: int = None) -> np.ndarray:
    """Calls image_add_text to add title on an updated image with padding on top for space and text centered"""

    assert len(image.shape) == 3, f"Wrong image shape: {image.shape}"
    height, _ = image.shape[0:2]
    pil_image = Image.fromarray(image.astype(np.uint8))
    draw = ImageDraw.Draw(pil_image)

    if size_px is None:
        size_px = int(height * 0.10)
        logger.debug2(f"Size of the font not provided. Giving 10% of the image = {size_px}")

    font = font or _get_default_font(size_px=size_px)
    text_width, text_height = _pil_image_draw_textsize(draw, text, font)
    logger.debug2(f"Font: {font}")

    top_padding = top_padding or text_height
    if top_padding < text_height:
        logger.warning(f"{top_padding=} is below {text_height=} of {size_px=} {text=}.")

    # Expand the image with (left=0, top=top_padding, right=0, bottom=0)
    border = (0, top_padding, 0, 0)
    expanded_image = ImageOps.expand(pil_image, border=border, fill=background_color)
    expanded_image = np.array(expanded_image)
    if text_width > expanded_image.shape[1]:
        logger.warning(f"{text_width=} of {text=} is larger than {expanded_image.shape=}")

    position = (top_padding - text_height) // 2, (expanded_image.shape[1] - text_width) // 2
    return pil_image_add_text(expanded_image, text=text, position=position, font=font, font_color=font_color)

def pil_image_add_text(image: np.ndarray, text: str, position: tuple[int, int], font: str | None = None,
                       font_size_px: int = None, font_color: str = "white") -> np.ndarray:
    """Adds a text to the image"""
    assert len(image.shape) == 3, f"Wrong image shape: {image.shape}"
    font = font if font is not None else _get_default_font(font_size_px)
    height, width = image.shape[0:2]
    if not 0 <= position[0] <= height:
        logger.debug2(f"Height position ({position[0]}) is negative or outside boundaries ({height=}).")
    if not 0 <= position[1] <= width:
        logger.debug2(f"Width position ({position[1]}) is negative or outside boundaries ({width=}).")
    pil_image = Image.fromarray(image.copy())
    draw = ImageDraw.Draw(pil_image)
    # position (h, w) => draw.tetxt((w, h), ...)
    draw.text(position[::-1], text, font=font, fill=font_color)
    new_image = np.array(pil_image, dtype=image.dtype)
    return new_image

def pil_image_read(path: str) -> np.ndarray:
    """PIL image reader"""
    # TODO: for grayscale, this returns a RGB image too
    img_pil = Image.open(path)
    img_np = np.array(img_pil, dtype=np.uint8)
    # grayscale -> 3 gray channels repeated.
    if img_pil.mode == "L":
        return np.repeat(img_np[..., None], 3, axis=-1)
    # RGB or RGBA
    return img_np[..., 0:3]

def pil_image_write(file: np.ndarray, path: str):
    """PIL image writer"""
    assert file.min() >= 0 and file.max() <= 255
    img = Image.fromarray(file.astype(np.uint8), "RGB")
    img.save(path)

def _get_default_font(size_px: int | None = None):
    """
    Gets the default font for some pixel size. Downloads it if it's not provided in:
        '../../../resources/OpenSans-Bold.ttf'
    """
    default_font_heights = {12: 2, 2: 1, 3: 1, 4: 1, 5: 1, 6: 1, 7: 1, 8: 1, 9: 1, 10: 1, 11: 1, 13: 4, 14: 5, 15: 7, 16: 8, 17: 10, 18: 11, 19: 12, 20: 14, 21: 15, 22: 17, 23: 18, 24: 19, 25: 21, 26: 22, 27: 24, 28: 25, 29: 26, 30: 28, 31: 29, 32: 31, 33: 32, 34: 33, 35: 35, 36: 36, 37: 38, 38: 39, 39: 40, 40: 42, 41: 44, 42: 45, 43: 46, 44: 48, 45: 49, 46: 51, 47: 52, 48: 53, 49: 55, 50: 56, 51: 58, 52: 59, 53: 60, 54: 62, 55: 63, 56: 65, 57: 66, 58: 67, 59: 69, 60: 70, 61: 72, 62: 73, 63: 74, 64: 76, 65: 77, 66: 79, 67: 80, 68: 81, 69: 83, 70: 84, 71: 86, 72: 87, 73: 88, 74: 90, 75: 91, 76: 93, 77: 94, 78: 95, 79: 97, 80: 98, 81: 100, 82: 101, 83: 102, 84: 104, 85: 105, 86: 107, 87: 108, 88: 109, 89: 111, 90: 112, 91: 114, 92: 115, 93: 116, 94: 118, 95: 119, 96: 121, 97: 122, 98: 123, 99: 125, 100: 126, 101: 128, 102: 129, 103: 130, 104: 132, 105: 133, 106: 135, 107: 136, 108: 137, 109: 139, 110: 140, 111: 142, 112: 143, 113: 144, 114: 146, 115: 147, 116: 149, 117: 150, 118: 151, 119: 153, 120: 154, 121: 156, 122: 157, 123: 158, 124: 160, 125: 161, 126: 163, 127: 164, 128: 165, 129: 167, 130: 168, 131: 170, 132: 171, 133: 172, 134: 174, 135: 175, 136: 177, 137: 178, 138: 179, 139: 181, 140: 182, 141: 184, 142: 185, 143: 186, 144: 188, 145: 189, 146: 191, 147: 192, 148: 194, 149: 195, 150: 196, 151: 198, 152: 199, 153: 201, 154: 202, 155: 203, 156: 205, 157: 206, 158: 208, 159: 209, 160: 210, 161: 212, 162: 213, 163: 215, 164: 216, 165: 217, 166: 219, 167: 220, 168: 222, 169: 223, 170: 224, 171: 226, 172: 227, 173: 229, 174: 230, 175: 231, 176: 233, 177: 234, 178: 236, 179: 237, 180: 238, 181: 240, 182: 241, 183: 243, 184: 244, 185: 245, 186: 247, 187: 248, 188: 250, 189: 251, 190: 252, 191: 254, 192: 255, 193: 257, 194: 258, 195: 259, 196: 261, 197: 262, 198: 264, 199: 265, 200: 266, 201: 268, 202: 269, 203: 271, 204: 272, 205: 273, 206: 275, 207: 276, 208: 278, 209: 279, 210: 280, 211: 282, 212: 283, 213: 285, 214: 286, 215: 287, 216: 289, 217: 290, 218: 292, 219: 293, 220: 294, 221: 296, 222: 297, 223: 299, 224: 300, 225: 301, 226: 303, 227: 304, 228: 306, 229: 307, 230: 308, 231: 310, 232: 311, 233: 313, 234: 314, 235: 315, 236: 317, 237: 318, 238: 320, 239: 321, 240: 322, 241: 324, 242: 325, 243: 327, 244: 328, 245: 329, 246: 331, 247: 332, 248: 334, 249: 335, 250: 336, 251: 338, 252: 339, 253: 341, 254: 342, 255: 343, 256: 345, 257: 346, 258: 348, 259: 349, 260: 350, 261: 352, 262: 353, 263: 355, 264: 356, 265: 357, 266: 359, 267: 360, 268: 362, 269: 363, 270: 364, 271: 366, 272: 367, 273: 369, 274: 370, 275: 371, 276: 373, 277: 374, 278: 376, 279: 377, 280: 378, 281: 380, 282: 381, 283: 383, 284: 384, 285: 385, 286: 387, 287: 388, 288: 390, 289: 391, 290: 392, 291: 394, 292: 395, 293: 397, 294: 398, 295: 399, 296: 401, 297: 402, 298: 404, 299: 405, 300: 406, 301: 408, 302: 409, 303: 411, 304: 412, 305: 413, 306: 415, 307: 416, 308: 418, 309: 419, 310: 420, 311: 422, 312: 423, 313: 425, 314: 426, 315: 427, 316: 429, 317: 430, 318: 432, 319: 433, 320: 434, 321: 436, 322: 437, 323: 439, 324: 440, 325: 441, 326: 443, 327: 444, 328: 446, 329: 447, 330: 448, 331: 450, 332: 451, 333: 453, 334: 454, 335: 455, 336: 457, 337: 458, 338: 460, 339: 461, 340: 462, 341: 464, 342: 465, 343: 467, 344: 468, 345: 469, 346: 471, 347: 472, 348: 474, 349: 475, 350: 476, 351: 478, 352: 479, 353: 481, 354: 482, 355: 483, 356: 485, 357: 486, 358: 488, 359: 489, 360: 490, 361: 492, 362: 493, 363: 495, 364: 496, 365: 497, 366: 499, 367: 500} # pylint: disable=line-too-long
    if size_px is None:
        size_px = 12
    assert isinstance(size_px, int)
    if size_px not in default_font_heights:
        logger.info(f"Size px {size_px} not in default_fonts")
        if size_px > max(default_font_heights.keys()):
            size_px = max(default_font_heights.keys())
        else:
            assert False
    font_path = get_project_root() / "resources/OpenSans-Bold.ttf"
    if not font_path.exists():
        font_path.parent.mkdir(exist_ok=True, parents=True)
        with open(font_path, "wb") as file:
            import requests # pylint: disable=import-outside-toplevel
            file.write(requests.get("https://github.com/edx/edx-fonts/raw/refs/heads/master/open-sans/fonts/Bold/OpenSans-Bold.ttf").content) # pylint: disable=all
    logger.debug2(f"Getting default font from '{font_path}' for desired height = '{size_px}' px")
    size = default_font_heights[size_px]
    if size == 0:
        logger.debug2("Asking for a size that is too small. Defaulting to font size = 1.")
        size = 1
    font = ImageFont.truetype(str(font_path), size=default_font_heights[size_px])
    return font

def _pil_image_draw_textsize(draw: ImageDraw, text: str, font: ImageFont) -> tuple[int, int]:
    im = Image.new(mode="P", size=(0, 0))
    draw = ImageDraw.Draw(im)
    _, _, width, _ = draw.textbbox((0, 0), text=text, font=font)
    _, _, _, height = draw.textbbox((0, 0), text="g", font=font) # g has highest width
    return width, height



"""setup.py -- note use setuptools==73.0.1; older versions fuck up the data files, newer versions include resources."""
from pathlib import Path
from setuptools import setup, find_packages

NAME = "video-representations-extractor"
VERSION = "1.16.0"
DESCRIPTION = "Video Representations Extractor (VRE) for computing algorithmic or neural representations of each frame."
URL = "https://gitlab.com/video-representations-extractor/video-representations-extractor"

CWD = Path(__file__).absolute().parent
with open(CWD/"README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

REQUIRED_CORE = [
    "vre-video>=0.5.0",
    "loggez>=0.8.4",
    "numpy>=1.21.6",
    "PyYAML==6.0.3",
    "tqdm==4.66.5",
    "overrides==7.7.0",
    "Pillow==11.3.0",
    "graphviz==0.20.3",
]

REQUIRED_REPOSITORY = [
    "torch==2.8.0",
    "torchvision==0.23.0",
    "opencv-python==4.12.0.88",
    "pycocotools==2.0.10",
    "timm==1.0.9",
    "diffusers==0.30.3",
]

def _filter_file(x: Path) -> bool:
    return (x.is_file() and x.suffix not in (".py", ".pyc", ".png", ".jpg", "*.md")
            and x.name != ".gitignore" and "weights" not in x.parts)
glob_files = lambda x: list(Path(x).glob("**/*")) # pylint: disable=all
vre_files = [str(x) for x in glob_files("vre/") if _filter_file(x)]
vre_repo_files = [str(x) for x in glob_files("vre_repository/") if _filter_file(x)]
data_files = [("", [*vre_files, *vre_repo_files])]

packages = find_packages()

setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=URL,
    packages=packages,
    data_files=data_files,
    package_data={"": data_files[0][1]},
    include_package_data=False,
    install_requires=REQUIRED_CORE,
    extras_require={
        "core": [],
        "repository": REQUIRED_REPOSITORY,
    },
    dependency_links=[],
    license="MIT",
    python_requires=">=3.10",
    scripts=["cli/vre", "cli/vre_collage", "cli/vre_reader", "cli/vre_gpu_parallel", "cli/vre_dir_analysis"],
)



"""
Helper mixin class that adds the weights/device relevant methods & properties for such representation
A representation that inherits this also must have weights in the weights repository.
"""
from abc import abstractmethod, ABC

from vre.logger import vre_logger as logger

try:
    import torch as tr
    VREDevice = str | tr.device # not only torch, but this is what we support atm
except ImportError:
    VREDevice = str

class LearnedRepresentationMixin(ABC):
    """Learned Representastion Mixin for VRE implementation"""
    def __init__(self):
        self._device: VREDevice | None = None
        self.setup_called = False

    @abstractmethod
    def vre_setup(self, load_weights: bool = True):
        """
        Setup method for this representation. This is required to run this representation from within VRE. We do this
        setup separately, so we can instatiate the object without doing any VRE specific setup, like loading weights.
        """

    @abstractmethod
    def vre_free(self):
        """Needed to deallocate stuff from cuda mostly. After this, you need to run vre_setup() again."""

    @staticmethod
    @abstractmethod
    def get_weights_paths(variant: str | None = None) -> list[str]:
        """Returns the list of paths to the weights of this model"""

    @property
    def device(self) -> VREDevice:
        """Returns the device of the representation"""
        if self._device is None:
            logger.warning(f"[{self}] No device set, returning 'cpu'. Call set_learned_params")
            return "cpu"
        return self._device

    @device.setter
    def device(self, dev: VREDevice):
        assert isinstance(dev, str) or (hasattr(dev, "type") and isinstance(dev.type, str)), dev
        self._device = dev

    def set_learned_params(self, **kwargs):
        """set the learned parameters for the representation"""
        attributes = ["device"]
        res = ""
        for attr in attributes:
            if attr in kwargs:
                setattr(self, attr, kwargs[attr])
                res += f"\n-{attr}: {kwargs[attr]}"

        if len(res) > 0:
            logger.debug(f"[{self}] Set node specific 'Learned' params:{res}")
